@@ -30,7 +30,7 @@
  * 
  *  Add total focused time to the grouping score function
  *  Fine tune the grouping scoring function
- *  Fix "changes" to naming and icon for groups/applications 
+ *  Fix "changes" to icon for groups/applications 
  *  Maybe fine tune size_allocate() some more...
  *  Better vertical layout handling
  *  prefs
@@ -79,6 +79,7 @@ struct _WnckTask
   gulong state_changed_tag;
   gulong icon_changed_tag;
   gulong name_changed_tag;
+  gulong app_name_changed_tag;
 
   GtkWidget *menu;
 };
@@ -147,6 +148,7 @@ static void     wnck_tasklist_forall        (GtkContainer     *container,
                                              gpointer          callback_data);
 static void     wnck_tasklist_remove	    (GtkContainer   *container,
 					     GtkWidget	    *widget);
+static void     wnck_tasklist_free_tasks    (WnckTasklist   *tasklist);
 static void     wnck_tasklist_update_lists  (WnckTasklist   *tasklist);
 static int      wnck_tasklist_layout        (GtkAllocation  *allocation,
 					     int             max_width,
@@ -271,6 +273,13 @@ wnck_task_finalize (GObject *object)
       task->name_changed_tag = 0;
     }
 
+  if (task->app_name_changed_tag)
+    {
+      g_signal_handler_disconnect (task->application,
+				   task->app_name_changed_tag);
+      task->app_name_changed_tag = 0;
+    }
+  
   if (task->menu)
     {
       gtk_widget_destroy (task->menu);
@@ -364,6 +373,13 @@ wnck_tasklist_finalize (GObject *object)
 
   tasklist = WNCK_TASKLIST (object);
 
+  /* Tasks should have gone away due to removing their
+   * buttons in container destruction
+   */
+  g_assert (tasklist->priv->windows == NULL);
+  g_assert (tasklist->priv->applications == NULL);
+  /* wnck_tasklist_free_tasks (tasklist); */
+  
   g_hash_table_destroy (tasklist->priv->win_hash);
   tasklist->priv->win_hash = NULL;
   
@@ -766,11 +782,10 @@ wnck_tasklist_forall (GtkContainer *container,
 
 static void
 wnck_tasklist_remove (GtkContainer   *container,
-		      GtkWidget	    *widget)
+		      GtkWidget	     *widget)
 {
   WnckTasklist *tasklist;
   GList *tmp;
-  gboolean found = FALSE;
   
   g_return_if_fail (WNCK_IS_TASKLIST (container));
   g_return_if_fail (widget != NULL);
@@ -790,15 +805,15 @@ wnck_tasklist_remove (GtkContainer   *container,
 	  tasklist->priv->windows =
 	    g_list_remove (tasklist->priv->windows,
 			   task);
-	  gtk_widget_destroy (task->button);
-	  found = TRUE;
+
+          gtk_widget_unparent (widget);
+          g_object_unref (task);
 	  break;
 	}
-      
     }
   
   tmp = tasklist->priv->applications;
-  while (!found && (tmp != NULL))
+  while (tmp != NULL)
     {
       WnckTask *task = WNCK_TASK (tmp->data);
       tmp = tmp->next;
@@ -810,12 +825,12 @@ wnck_tasklist_remove (GtkContainer   *container,
 	  tasklist->priv->applications =
 	    g_list_remove (tasklist->priv->applications,
 			   task);
-	  gtk_widget_destroy (task->button);
+
+          gtk_widget_unparent (widget);
+          g_object_unref (task);
 	  break;
 	}
     }
-
-  gtk_widget_unparent (widget);
 }
 
 
@@ -851,16 +866,10 @@ wnck_tasklist_new (WnckScreen *screen)
 }
 
 static void
-wnck_tasklist_update_lists (WnckTasklist *tasklist)
+wnck_tasklist_free_tasks (WnckTasklist *tasklist)
 {
-  GList *windows;
-  WnckWindow *win;
-  WnckApplication *app;
   GList *l;
-  WnckWorkspace *active_workspace;
-  WnckTask *app_task;
-  WnckTask *win_task;
-
+  
   tasklist->priv->active_task = NULL;
   tasklist->priv->active_app = NULL;
   
@@ -870,8 +879,11 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
       while (l != NULL)
 	{
 	  WnckTask *task = WNCK_TASK (l->data);
-	  l = l-> next;
-	  g_object_unref (task);
+	  l = l->next;
+          /* if we just unref the task it means we lose our ref to the
+           * task before we unparent the button, which breaks stuff.
+           */
+	  gtk_widget_destroy (task->button);
 	}
     }
   g_assert (tasklist->priv->windows == NULL);
@@ -883,13 +895,30 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
       while (l != NULL)
 	{
 	  WnckTask *task = WNCK_TASK (l->data);
-	  l = l-> next;
-	  g_object_unref (task);
+	  l = l->next;
+          /* if we just unref the task it means we lose our ref to the
+           * task before we unparent the button, which breaks stuff.
+           */
+	  gtk_widget_destroy (task->button);
 	}
     }
   g_assert (tasklist->priv->applications == NULL);
   g_assert (g_hash_table_size (tasklist->priv->app_hash) == 0);
+}
 
+static void
+wnck_tasklist_update_lists (WnckTasklist *tasklist)
+{
+  GList *windows;
+  WnckWindow *win;
+  WnckApplication *app;
+  GList *l;
+  WnckWorkspace *active_workspace;
+  WnckTask *app_task;
+  WnckTask *win_task;
+
+  wnck_tasklist_free_tasks (tasklist);
+  
   active_workspace = wnck_screen_get_active_workspace (tasklist->priv->screen);  
   windows = wnck_screen_get_windows (tasklist->priv->screen);
   
@@ -1207,9 +1236,10 @@ wnck_task_popup_menu (WnckTask  *task)
       
       gtk_widget_show (menu_item);
       
-      g_signal_connect (G_OBJECT (menu_item), "activate",
-			G_CALLBACK (wnck_task_menu_activated), win_task);
-      
+      g_signal_connect_object (G_OBJECT (menu_item), "activate",
+                               G_CALLBACK (wnck_task_menu_activated),
+                               G_OBJECT (win_task),
+                               0);      
       
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
       
@@ -1263,8 +1293,9 @@ wnck_task_get_text (WnckTask *task)
   
   if (task->is_application)
     {
-      /* FIXME: What if the application name changes */
-      return g_strdup (wnck_application_get_name (task->application));
+      return g_strdup_printf ("%s (%d)",
+                              wnck_application_get_name (task->application),
+                              wnck_application_get_n_windows (task->application));
     }
   else
     {
@@ -1457,6 +1488,16 @@ wnck_task_name_changed (WnckWindow *window,
     wnck_task_update_visible_state (task);
 }
 
+static void
+wnck_task_app_name_changed (WnckApplication *app,
+                            gpointer         data)
+{
+  WnckTask *task = WNCK_TASK (data);
+
+  if (task)
+    wnck_task_update_visible_state (task);
+}
+
 static gboolean
 wnck_task_button_press_event (GtkWidget	      *widget,
 			      GdkEventButton  *event,
@@ -1490,7 +1531,9 @@ wnck_task_create_widgets (WnckTask *task)
   char *text;
   
   task->button = gtk_toggle_button_new ();
-
+  g_object_add_weak_pointer (G_OBJECT (task->button),
+                             (void**) &task->button);
+  
   gtk_widget_set_name (task->button,
 		       "tasklist-button");
 
@@ -1531,19 +1574,28 @@ wnck_task_create_widgets (WnckTask *task)
   g_free (text);
   
   /* Set up signals */
-  g_signal_connect (G_OBJECT (task->button), "toggled",
-		    G_CALLBACK (wnck_task_button_toggled), task);
+  g_signal_connect_object (G_OBJECT (task->button), "toggled",
+                           G_CALLBACK (wnck_task_button_toggled),
+                           G_OBJECT (task),
+                           0);
 
   
-  g_signal_connect (G_OBJECT (task->button), "button_press_event",
-		    G_CALLBACK (wnck_task_button_press_event), task);
+  g_signal_connect_object (G_OBJECT (task->button), "button_press_event",
+                           G_CALLBACK (wnck_task_button_press_event),
+                           G_OBJECT (task),
+                           0);
 
-  if (!task->is_application)
+  if (task->is_application)
+    {
+      task->app_name_changed_tag = g_signal_connect (G_OBJECT (task->application), "name_changed",
+                                                     G_CALLBACK (wnck_task_app_name_changed), task);
+    }
+  else
     {
       task->state_changed_tag = g_signal_connect (G_OBJECT (task->window), "state_changed",
-						  G_CALLBACK (wnck_task_state_changed), task->tasklist);
+                                                  G_CALLBACK (wnck_task_state_changed), task->tasklist);
       task->icon_changed_tag = g_signal_connect (G_OBJECT (task->window), "icon_changed",
-						 G_CALLBACK (wnck_task_icon_changed), task);
+                                                 G_CALLBACK (wnck_task_icon_changed), task);
       task->name_changed_tag = g_signal_connect (G_OBJECT (task->window), "name_changed",
 						 G_CALLBACK (wnck_task_name_changed), task);
     }
@@ -1650,8 +1702,10 @@ wnck_task_new_from_application (WnckTasklist    *tasklist,
   
   wnck_task_create_widgets (task);
   
-  g_signal_connect_after (task->button, "expose_event",
-			  G_CALLBACK (wnck_task_app_expose), task);
+  g_signal_connect_object (task->button, "expose_event",
+                           G_CALLBACK (wnck_task_app_expose),
+                           G_OBJECT (task),
+                           G_CONNECT_AFTER);
   
   return task;
 }
